
package require struct::list

oo::class create cnt::db {

	variable id
	variable channel

	constructor {args} {
		variable irc [lindex $args 0]
		variable channel [lindex $args 1]
		if {[llength $args] > 2} {
			my construct {*}[lrange $args 2 end]
		}
	}

	method type {} {
		info object class [self object]
	}

	method save {} {
		variable irc
		set data [my tolist]
		variable id [$irc db Save-[lindex $data 0] [lrange $data 1 end]]
	}

	method update {} {
		if {![info exists id]} {
			error "Trying to update an object which is not yet saved"
		}

		variable irc
		set data [my tolist]
		$irc db Update-[lindex $data 0] $id [lrange $data 1 end]
	}

	method id {value} {
		variable id
		set id $value
	}

	method irc {} {
		variable irc
		return $irc
	}

	method channel {} {
		variable channel
		return $channel
	}
}

::hlirc::define-module cnt-db {

	module dispatcher cmd-db DB

	proc do-connecting {} {
		set filename [file join [module datadir] cnt.db]
		module log info "Using database $filename"

		if {[info procs db] == ""} {
			sqlite3 db $filename

			db eval {
				CREATE TABLE IF NOT EXISTS votings (
					id integer PRIMARY KEY AUTOINCREMENT,
					type text,
					line text,
					created integer,
					threshold text,
					votes text,
					result text,

					channel text
				)
			}

			db eval {
				CREATE TABLE IF NOT EXISTS delegates (
					user text,
					cfrom text,
					cto text
				)
			}

			db eval {
				CREATE TABLE IF NOT EXISTS users (
					user text,

					channel text
				)
			}

			db eval {
				CREATE TABLE IF NOT EXISTS plenaries (
					type text,
					delegates integer,
					channels text,
					plenarychannel text UNIQUE,
					created integer,
					power text,
					afk text
				)
			}

			db eval {
				CREATE TABLE IF NOT EXISTS plenarypassword (
					plenarychannel text,
					password text
				)
			}
		}
	}

	# Votings

	proc load-voting {data} {
		set voting [[lindex $data 1] new [irc] [lindex $data end]]
		$voting id [lindex $data 0]
		$voting fromlist [lrange $data 2 end]
		return $voting
	}

	proc DB-Save-voting {data} {
		lassign $data type line created threshold votes result channel
		db eval "INSERT INTO votings(channel, type, line, created, threshold, votes, result)
							VALUES (:channel, :type, :line, :created, :threshold, :votes, :result)"
		return [db last_insert_rowid]
	}

	proc DB-Update-voting {id data} {
		set votes [lindex $data 4]
		set result [lindex $data 5]
		db eval "UPDATE votings SET votes=:votes, result=:result WHERE id=:id"
	}

	proc DB-select-voting {id} {
		set data [db eval "SELECT * FROM votings WHERE id=:id LIMIT 1"]
		if {$data == ""} {
			throw {invalid voting} "There's no voting with id $id"
		}
		load-voting $data
	}

	proc DB-pending-votings {channel} {
		set data [db eval "SELECT * FROM votings WHERE result == \"\" AND channel == :channel"]
		set data [group $data 7]
		struct::list map $data load-voting
	}

	proc DB-list-votings {channel page} {
		incr page -1
		set data [db eval "SELECT * FROM votings WHERE channel == :channel ORDER BY created DESC LIMIT 10 OFFSET :page * 10"]
		set data [group $data 8]
		struct::list map $data load-voting
	}

	# Delegates

	proc DB-delegates {args} {
		cnt::named $args {user "" from "" to ""}
		set user [hlirc::canonicalize $user]
		set from [hlirc::canonicalize $from]
		set to [hlirc::canonicalize $to]
		set data [db eval "SELECT * FROM delegates WHERE (cfrom = :from OR :from = '') \
			AND (cto = :to OR :to = '') \
			AND (user = :user OR :user = '') "]
		group $data 3
	}

	proc DB-remove-delegates {args} {
		cnt::named $args {user "" from "" to ""}
		if {$from == "" && $user == "" && $to == ""} {
			throw {db options} "At least one argument is required"
		}
		db eval "DELETE FROM delegates WHERE (cfrom = :from OR :from = '') \
			AND (cto = :to OR :to = '') \
			AND (user = :user OR :user = '') "
	}

	proc DB-register-delegate {user from to} {
		if {[DB-delegates -user $user -from $from -to $to] != ""} {
			module log error "Delegate $user from $from to $to already registered"
			return
		}
		db eval "INSERT INTO delegates VALUES (:user, :from, :to)"
	}

	# Users

	proc DB-register-user {channel user} {
		db eval "INSERT INTO users VALUES (:user, :channel)"
	}

	proc DB-unregister-user {channel user} {
		db eval "DELETE FROM users WHERE user = :user AND channel = :channel"
	}

	proc DB-user? {channel user} {
		db eval "SELECT count(*) FROM users WHERE user = :user AND channel = :channel"
	}

	proc DB-users {channel} {
		db eval "SELECT user FROM users WHERE channel = :channel"
	}

	# Plenaries
	
	proc DB-new-plenary {type delegates channels plenarychannel} {
		set channels [struct::list map $channels hlirc::canonicalize]
		set plenarychannel [hlirc::canonicalize $plenarychannel]
		set password [cnt::random-string 20]
		set created [clock seconds]
		db eval "INSERT INTO plenaries VALUES (:type, :delegates, :channels, :plenarychannel, :created, 'same', 'split')"
		db eval "INSERT INTO plenarypassword VALUES (:plenarychannel, :password)"
	}

	proc DB-plenary-password {plenarychannel} {
		db eval "SELECT password FROM plenarypassword WHERE plenarychannel = :plenarychannel"
	}

	proc DB-all-plenarychannels {} {
		db eval "SELECT plenarychannel FROM plenaries"
	}

	proc DB-select-plenary {plenarychannel} {
		set plenarychannel [hlirc::canonicalize $plenarychannel]
		set data [db eval "SELECT * FROM plenaries WHERE plenarychannel = :plenarychannel"]
		if {$data != ""} {
			make-plenary-dict $data
		} else {
			return ""
		}
	}

	# Returns all plenaries the channel is invited and we host
	proc DB-channel-plenaries {channel} {
		set data [db eval "SELECT * FROM plenaries WHERE channels LIKE '%${channel}%'"]
		return [struct::list map [group $data 7] make-plenary-dict]
	}

	proc DB-delete-plenary {plenarychannel} {
		db eval "DELETE FROM plenaries WHERE plenarychannel = :plenarychannel"
		db eval "DELETE FROM plenarypassword WHERE plenarychannel = :plenarychannel"
	}

	proc make-plenary-dict {row} {
		dict create type [lindex $row 0] \
					delegates [lindex $row 1] \
					channels [lindex $row 2] \
					plenarychannel [lindex $row 3] \
					created [lindex $row 4] \
					power [lindex $row 5] \
					afk [lindex $row 6]
	}

	# Util
	proc group {list num} {
		set result [list]
		set row [list]
		for {set i 0} {$i < [llength $list]} {incr i} {
			if {$i % $num == 0 && $i != 0} {
				lappend result $row
				set row [list]
			}
			lappend row [lindex $list $i]
		}
		if {$row != {}} {
			lappend result $row
		}
		return $result
	}
}
