
package require struct::list

::hlirc::define-module cnt {

	module require state
	module require services
	module require cnt-db
	module require ibc

	module config threshold majority {How many votes are required to decide on the voting.
								majority/unanimous/second}
	module config default-vote abstain {All who didn't vote before the timeout are considered
									  to have voted this. For votings with non-default
									  options it's always abstain.}
	module config timeout 60 {How much time there is to vote. In seconds.}
	module config security none {Security for bot's channel. Can be none, poll or password <password>}
	module config public-polls no {Can users from other channels request polls data.}
	module config public-settings no {Can users from other channels request settings data.}
	module config public-delegates no {Can users from other channels request delegates data.}

	module config mainchannel "" {The channel this bot is going to serve}
	module config mainchannel-password "" {The password used to register and access the main channel.
										WARNING: If it's not empty the bot has probably already registered the channel with this password.
										Don't change.}

	module wizard {
		puts "Please enter the channel this bot is going to serve"
		set mainchannel [::hlirc::wizard query "Channel name:" "" ::hlirc::channel? "This doesn't look like a channel name."]
		module config mainchannel $mainchannel
	}

	proc cmd-mainchannel {} {
		::hlirc::canonicalize [module config mainchannel]
	}

	variable init true

	proc do-connected {} {
		variable init true
		module log notice "Finished connecting to the server"
		set plenaries [[irc] db all-plenarychannels]
		if {$plenaries != ""} {
			set left [llength $plenaries]
			# Rejoin all plenary channels
			foreach channel $plenaries {
				[irc] log info "Checking plenaries. $left left..."
				[irc] join $channel
				module after 2000
				incr left -1
			}
			module after 5000
		}
		set init false

		[irc] log notice "Finished checking plenary channels. Joining the main channel"
		[irc] join [[irc] mainchannel]
		module after idle normalize-plenaries
	}

	proc do-joined-channel {channel} {
		variable init
		set chan [[irc] mainchannel]
		if {$channel == $chan} {
			[irc] services-op $chan
		}

		set plenary [[irc] db select-plenary $channel]
		if {$plenary != ""} {
			# Joined a plenary channel
			if {$init} {
				# During initialization
				set users [channel-users $channel]
				if {[llength $users] == 0} {
					# And it's empty
					if {([clock seconds] - [dict get $plenary created]) > 600} {
						# And older than 10 minutes
						dismiss-plenary $plenary
					}
				} else {
					[irc] mode $channel +i
				}
			} else {
				# Joined a new plenary channel
				# Register it
				[irc] register-channel $channel [[irc] db plenary-password $channel]
				# And make it invite-only
				[irc] mode $channel +i
			}
		}
	}

	proc normalize-plenaries {} {
		set registered [hlirc::canonicalize [[irc] list-registered-channels]]
		set needed [hlirc::canonicalize [concat [[irc] mainchannel] [[irc] db all-plenarychannels]]]
		set extrachannels [struct::set difference $registered $needed]
		set extraplenaries [struct::set difference $needed $registered]
		if {$extrachannels != ""} {
			module log info "Dropping extra channels: $extrachannels"
		}
		if {$extraplenaries != ""} {
			module log info "Dropping extra plenaries: $extraplenaries"
		}
		foreach todrop $extrachannels {
			[irc] unregister-channel $todrop "" ;# Some services require password :/
			module after 2000
		}
		foreach todrop $extraplenaries {
			[irc] db delete-plenary $todrop
		}
	}

	proc do-unregistered-channel {channel} {
		# The ChanServ told us that the main channel is unregistered
		if {$channel == [[irc] mainchannel]} {
			# Make a password if the config doesn't have one
			if {[module config mainchannel-password] == ""} {
				module config mainchannel-password [cnt::random-string 20]
			}
			# And register the channel
			[irc] register-channel $channel [module config mainchannel-password]
		}
	}

	proc do-kick {} {
		set nick [additional 0]
		set user [[irc] nick-user $nick]
		set channel [target]
		set reason [msg]
		if {$reason == ""} {
			set reason "Kicked from $channel"
		}
		if {$user != ""} {
			remove-delegate $channel $nick $reason
			[irc] db unregister-user $channel $nick
		}
	}

	proc do-part {} {
		if {[who] == [[irc] state nick]} return
		check-plenary [target]
	}

	proc do-quit {} {
		set plenaries [[irc] db all-plenarychannels]
		struct::list map $plenaries check-plenary
	}

	proc check-plenary {channel} {
		if {[[irc] db select-plenary $channel] != ""} {
			set users [channel-users $channel]
			if {[llength $users] <= 0} {
				set plenary [[irc] db select-plenary $channel]
				dismiss-plenary $plenary
			}
		}
	}

	command start {
		!start type topic
	} {
		variable init
		if {$init} return
		match {type {rest rest}} {
			if {$type in [list approve-delegate attend-plenary]} {
				[irc] reply "Error: This type of voting can't be started manually"
			}
			try {
				if {[hlirc::channel? [target]]} {
					create-voting $type [target] [join $rest]
				} else {
					[irc] reply "Votings should be created on channels"
				}
			} trap {invalid voting} msg {
				[irc] reply "Error: $msg"
			}
		}
	}

	proc create-voting {type channel line} {
		set validator [namespace code [list validator $type $channel %v]]
		set voting [::cnt::voting $type [irc] $channel $line [module config threshold] $validator]
		[irc] privmsg $channel "Started a voting with id [$voting get id]: [$voting summary]"
		$voting destroy
	}

	proc validator {type channel voting} {
		switch $type {
			opcommand {
				if {[lindex [%v opcommand] 0] == "/mode" &&
					[string first "i" [lrange [%v opcommand] 1 end]]} {
					throw {invalid voting} "Please use the security setting to set/unset invisible flag"
				}
			}
			send-delegate {
				set otherchannel [$voting get otherchannel]
				set bot [[irc] ibc-find-bot $otherchannel]
				if {$bot == ""} {
					throw {invalid voting} "No known bot serves channel $otherchannel"
				}
				if {[$voting default-options?]} {
					check-delegate [$voting get delegate] $channel $otherchannel
				} else {
					foreach nick [$voting get options] {
						if {$nick ni {abstain nobody}} {
							check-delegate $nick $channel $otherchannel
						}
					}
				}
			}
			plenary {
				set channels [$voting get channels]
				foreach channel $channels {
					if {[[irc] ibc-find-bot $channel] == ""} {
						throw {invalid voting} "Can't find a bot serving channel $channel"
					}
				}
				if {![[irc] state channel-empty? $channel]} {
					throw {invalid voting plenary channel} "The channel $channel is already occupied. Please, try another one."
				}
			}
			recall-delegate {
				set nick [$voting get nick]
				set otherchannel [$voting get otherchannel]
				if {$nick != ""} {
					set user [[irc] nick-user $nick]
					if {$user == ""} {
						throw {invalid voting} "$nick is not logged in"
					}
				}
				if {[llength [[irc] db delegates -user $user -to $otherchannel]] == 0} {
					if {$nick == ""} {
						throw {invalid voting} "There are no delegates to channel $otherchannel"
					} elseif {$otherchannel == ""} {
						throw {invalid voting} "$nick is not delegated to any channel"
					} else {
						throw {invalid voting} "$nick is not delegated to $otherchannel"
					}
				}
			}
			default return
		}
	}

	proc check-delegate {nick from to} {
		if {![[irc] user-in-channel $nick $from]} {
			throw {invalid voting} "Unknown user $nick"
		}
		set user [[irc] nick-user $nick]
		if {$user == ""} {
			throw {invalid voting} "$nick is not logged in"
		}
		if {[[irc] db delegates -to $to -user $user] != ""} {
			throw {invalid voting} "User $user is already delegated to $to"
		}
		if {[[irc] db delegates -from $to -user $user] != ""} {
			throw {invalid voting} "User $user is a delegate FROM $to"
		}
	}

	command vote {
		!vote id vote
	} {
		variable init
		if {$init} return
		match {id vote} {
			try {
				set voting [[irc] db select-voting $id]
				set chan [$voting channel]
				if {![cmd-user-in-channel [who] $chan]} {
					[irc] reply "You need to be on $chan to vote"
					return
				}

				if {[$voting get result] != ""} {
					[irc] reply "This voting has already finished"
					return
				}

				set user [[irc] nick-user [who]]
				if {$user == ""} {
					[irc] reply "You need to be a registered user to vote"
					return
				}

				$voting vote $user $vote
				[irc] reply "Vote accepted"
				set voters [channel-users $chan]
				set voterscnt [llength $voters]
				incr voterscnt -1
				if {[$voting plenary?]} {
					module after idle plenary-voting-done? $voting
				} elseif {[$voting done? $voterscnt]} {
					finish-voting $voting
				}
			} trap {invalid} msg {
				[irc] reply "Error: $msg"
			}
		}
	}

	proc plenary-voting-done? {voting} {
		set plenarychannel [$voting channel]
		set plenary [cmd-get-plenary $plenarychannel]
		switch [dict get $plenary type] {
			public {
				set voters [channel-users $chan]
				set voterscnt [llength $voters]
				if {[$voting done? $voterscnt]} {
					finish-voting $voting
				}
			}
			delegates {
				set channels [dict get $plenary channels]
				set groups {}
				set online [[irc] state list users $plenarychannel]
				set powermode [dict get $plenary power]
				set afk [dict get $plenary afk]
				foreach channel $channels {
					if {$powermode == "same"} {
						set power 1
					} elseif {$powermode == "proportional"} {
						set power [[irc] state channel-usercount $channel]
					}
					set delegates [[irc] db delegates -to $plenarychannel -from $channel]
					set delegates [struct::list mapfor d $delegates {lindex $d 0}]
					set onlinedelegates [struct::set intersect $delegates $online]
					lappend groups [list $power {*}$onlinedelegates]
				}

				set delegatesnum [dict get $plenary delegates]
				if {[$voting done-groups? $groups $delegatesnum $afk]} {
					finish-voting $voting
				}
			}
		}
	}

	proc finish-voting {voting {reason "Threshold passed"}} {
		set channel [$voting channel]
		[irc] privmsg $channel "Voting [$voting get id]: [$voting summary]"
		[irc] privmsg $channel "$reason. Result: [$voting get result]"

		finish-[namespace tail [info object class $voting]] $voting
		$voting destroy
	}

	proc finish-poll {voting} {}
	
	proc finish-setting {voting} {
		if {[$voting get result] != yes} return

		set chan [$voting channel]
		if {[$voting get setting] == "security"} {
			if {[$voting get value] == "none"} {
				[irc] mode $chan -i
			} else {
				[irc] mode $chan +i
			}
		}
		module config [$voting get setting] [$voting get value]
	}

	proc finish-opcommand {voting} {
		if {[$voting get result] != yes} return

		set chan [$voting channel]
		set type [lindex [$voting get opcommand] 0]
		set args [join [lrange [split [$voting get opcommand]] 1 end]]
		switch $type {
			/ban {
				[irc] ban $chan $args
			}
			/kick {
				[irc] kick $chan $args
			}
			/kban {
				[irc] ban $chan $args
				[irc] kick $chan $args
			}
			/mode {
				[irc] mode $chan $args
			}
			/topic {
				[irc] topic $chan $args
			}
		}
	}

	proc finish-invite {voting} {
		set nick [$voting get nick]
		set user [$voting get user]
		set chan [$voting channel]
		if {[$voting get result] == yes} {
			[irc] db register-user $chan $user
			[irc] invite $chan $nick
		} else {
			[irc] privmsg $nick "The channel $chan refused to invite you"
		}
	}

	# DELEGATES

	proc finish-send-delegate {voting} {
		if {[$voting result] in [list nobody draw]} return

		set channel [$voting channel]
		set otherchannel [$voting get otherchannel]
		set nick [$voting result]
		set user [$voting result-user]
		if {$user == ""} {
			[irc] privmsg $channel "User $nick is not logged in"
		}

		if {[[irc] db delegates -user $user -to $otherchannel] != ""} {
			[irc] privmsg $channel "User $nick is already delegated to $otherchannel"
		} elseif {[[irc] db delegates -user $user -from $otherchannel] != ""} {
			[irc] privmsg $channel "User $nick is a delegate FROM $otherchannel"
		} else {
			module after idle ibc-approve-delegate $nick $channel $otherchannel
		}
	}

	proc finish-request-delegate {voting} {
		if {[$voting result] == "yes"} {
			module after idle ibc-request-delegate [$voting get otherchannel] [$voting channel]
		}
	}

	proc finish-approve-delegate {voting} {
		if {[$voting result] == "yes"} {
			do-approve-delegate [$voting get nick] [$voting get otherchannel] [$voting channel]
		}
	}

	proc finish-recall-delegate {voting} {
		if {[$voting result] != "yes"} return
		set channel [$voting channel]
		set otherchannel [$voting get otherchannel]
		set nick [$voting get nick]
		set user [$voting get user]
		if {$user == ""} {
			[irc] privmsg $channel "User $nick is not logged in"
		}
		set delegates	[concat [[irc] db delegates -from $otherchannel -user $user] \
								[[irc] db delegates -to $otherchannel -user $user]]
		foreach delegate $delegates {
			lassign $delegate nick from to
			module after idle ibc-recall-delegate $nick $from $to "Recalled by $channel"
		}
		[irc] db remove-delegates -user $user -from $otherchannel
		[irc] db remove-delegates -user $user -to $otherchannel
	}

	proc do-approve-delegate {nick from to} {
		set user [[irc] nick-user $nick]
		if {$user == ""} {
			[irc] privmsg $to "$nick is not logged in"
			return
		}
		[irc] db register-delegate $user $from $to
		[irc] privmsg $nick "You have been approved as a delegate to channel $to"
		[irc] invite $to $nick
		module after idle ibc-delegate-approved $from $to $nick
	}

	proc remove-delegate {nick reason} {
		set user [[irc] nick-user $nick]
		set delegates [[irc] db delegates -user $user]
		set mychannels [[irc] state channels]
		foreach delegate $delegates {
			lassign $delegate n f t
			if {$f in $mychannels} {
				module after idle ibc-recall-delegate $n $f $t $reason
			} elseif {$t in $mychannels} {
				module after idle ibc-delegate-kicked $n $f $t $reason
			}
		}
		[irc] db remove-delegates -user $user
	}

	proc do-delegate-kicked {channel nick reason} {
		set user [[irc] nick-user $nick]
		set delegates [[irc] db delegates -user $user -to $channel]
		foreach delegate $delegates {
			lassign $delegate user from to
			[irc] privmsg $from "Delegate $nick was kicked from $to. Reason: $reason"
			[irc] db remove-delegates -user $user -to $to
		}
	}

	# PLENARIES

	proc finish-plenary {voting} {
		if {[$voting result] != "yes"} return
		set type [$voting get type]
		set delegates [$voting get delegates]
		set channels [$voting get channels]
		set plenarychannel [$voting get plenarychannel]
		[irc] db new-plenary $type $delegates [join $channels] $plenarychannel
		module after idle ibc-notify-plenary [$voting channel] [[irc] db select-plenary $plenarychannel]
		[irc] join $plenarychannel
		module after 5000
		set chan [$voting channel]
		switch $type {
			public {
				set users [channel-users $chan]
				module after 10000 do-attend-plenary $chan $plenarychannel $users
			}
			delegates {
				[irc] privmsg $chan "A plenary is assembled with the following channels: [join $channels {, }]"
				[irc] privmsg $chan "You may send up to $delegates delegates to this plenary. use !start send-delegate $plenarychannel: <delegate candidates...>"
			}
		}
	}

	proc finish-attend-plenary {voting} {
		if {[$voting result] in [list draw no]} return
		set plenarychannel [$voting get plenarychannel]
		set plenary [cmd-get-plenary $plenarychannel]
		set users [channel-users [$voting channel]]
		module after idle ibc-attend-plenary [dict get $plenary bot] $plenarychannel $users
	}

	proc do-attend-plenary {source channel nicks} {
		set plenary [[irc] db select-plenary $channel]
		if {[dict get $plenary type] == "public"} {
			foreach nick $nicks {
				[irc] db register-user $channel [[irc] nick-user $nick]
				[irc] invite $channel $nick
			}
		}
	}

	proc dismiss-plenary {plenary} {
		set plenarychannel [dict get $plenary plenarychannel]
		module log notice "Dismissing plenary $plenarychannel"
		foreach channel [struct::set difference [dict get $plenary channels] [[irc] state channels]] {
			module after idle ibc-plenary-dismissed $channel $plenarychannel
		}
		set password [[irc] db plenary-password $plenarychannel]
		[irc] db delete-plenary $plenarychannel
		[irc] unregister-channel $plenarychannel $password
		[irc] part $plenarychannel
	}

	proc finish-invite-channel {voting} {
		set channel [$voting get channel-to-invite]
		set plenary [[irc] db select-plenary $channel]
		module log notice "Inviting channel $channel to plenary [dict get $plenary plenarychannel]"
		ibc-notify-plenary $channel $plenarychannel
	}

	# SHOW

	command show {
		!show [channel] <polls|settings|plenaries|delegates|users|bots|bot|help> ...
	} {
		variable init
		if {$init} return
		match {{optvar channel ::hlirc::channel? [[irc] mainchannel]}
			   {var what show-category?}
			   {rest args}} {
			show-to [who] $what $channel {*}$args
		}
	}

	proc show-category? {cat} {
		expr {$cat in [list poll polls \
							bot bots \
							setting settings \
							delegate delegates \
							plenaries help]}
	}

	proc show-to {nick what channel args} {
		try {
			if {[string trimright $what s] in [list poll setting delegate]} {
				set what [string trimright $what s]s
			}

			if {$channel ni [[irc] state channels]} {
				module after idle ibc-show-remote $channel $nick $what $args
			} else {
				if {![cmd-user-in-channel $nick $channel]} {
					if {$what in [list bots plenaries]} {
						[irc] privmsg $nick "Error: This data is only available from your bot"
						return
					}
					if {$what ni [list bot help] &&
						![module config public-${what}]} {
						[irc] privmsg $nick "Error: This data is not public"
						return
					}
				}
				if {$channel != [[irc] mainchannel] &&
					$what in [list settings]} {
					[irc] privmsg $nick "Error: Plenaries share host channel's settings"
				}
				try {
					show-${what} $nick $channel {*}$args
				} trap {db} msg {
					[irc] privmsg $nick "Error: $msg"
				}
			}
		} trap {hlirc command syntax} msg {
			show-show-help $what
		}
	}

	proc show-polls {nick channel args} {
		if {[llength $args] == 0} {
			set type list
		} else {
			set type [lindex $args 0]
		}
		switch $type {
			details {
				set pollid [lindex $args 1]
				if {$pollid == ""} {
					throw {hlirc command syntax} ""
				}
				try {
					set votings [[irc] db select-voting $pollid]
					set mode long
				} trap {invalid voting} msg {
					[irc] privmsg $nick $msg
					return
				}
			}
			pending {
				set votings [[irc] db pending-votings $channel]
				set mode short
			}
			list {
				set page [lindex $args 1]
				if {![string is integer $page]} {
					[irc] privmsg $nick "Page number must be a... number"
					return
				} elseif {$page == ""} {
					set page 1
				} else {
					incr page
				}
				set votings [[irc] db list-votings $channel $page]
				set mode short
			}
		}
		if {[llength $votings] == 0} {
			[irc] privmsg $nick "No votings found"
		} else {
			struct::list map $votings "show-poll $nick $mode"
			struct::list mapfor voting $votings {$voting destroy}
		}
	}

	proc show-poll {nick mode voting} {
		array set fullresult [$voting full-result]
		if {[$voting get result] == ""} {
			set result " pending"
		} else {
			set result " = [$voting get result]"
		}
		if {$mode == "short"} {
			set lastline "#[$voting get id]: [$voting summary]. Result$result. Count:"
		} else {
			[irc] privmsg $nick "Voting #[$voting get id] started on [clock format [$voting get created]]: [$voting summary]"
			set lastline "Result$result. Count:"
		}
		set results ""
		if {[$voting default-options?]} {
			foreach option [list yes no abstain] {
				append results "[set option]:[set fullresult($option)]; " }
		} else {
			set options [struct::list shuffle [struct::set difference [$voting get options] abstain]]
			foreach option $options {
				append results "[set option]:[set fullresult($option)]; "
			}
			append results "abstain:[set fullresult(abstain)]; "
		}
		[irc] privmsg $nick "$lastline $results"
	}

	proc show-settings {nick args} {
		foreach setting [list threshold default-vote public-polls public-settings public-delegates] {
			show-setting $nick $setting [module config $setting]
		}
		show-setting $nick security [lindex [module config security] 0] ;# Remove the password if it's there
		show-setting $nick timeout [::cnt::seconds-interval [module config timeout]]
	}

	proc show-setting {nick key value} {
		[irc] privmsg $nick "$key - $value"
	}

	proc show-bot {nick channel args} {
		[irc] privmsg $nick "I'm the bot which serves [[irc] mainchannel]"
		set plenarychannels [[irc] db all-plenarychannels]
		if {$plenarychannels != ""} {
			[irc] privmsg $nick "And hosts the following plenaries: [join $plenarychannels]"
		}
	}

	proc show-bots {nick args} {
		set data [[irc] ibc-listing bots]
		foreach line $data {
			[irc] privmsg $nick "[lindex $line 0]: [join [lindex $line 1]]"
		}
	}

	proc show-delegates {nick_to hostchannel args} {
		set nick ""
		set channel ""
		if {[llength $args] > 2} {
			throw {hlirc command syntax} ""
		}
		foreach arg $args {
			if {[::hlirc::channel? $arg]} {
				set channel $arg
			} else {
				set nick $arg
			}
		}
		set user [[irc] nick-user $nick]
		set data [concat [[irc] db delegates -to $hostchannel] [[irc] db delegates -from $hostchannel]]
		if {[llength $data] == 0} {
			[irc] privmsg $nick_to "No delegates found"
			return
		}
		foreach delegate $data {
			lassign $delegate n f t
			[irc] privmsg $nick_to "${n}: delegated from $f to $t"
		}
	}

	proc show-plenaries {nick channel} {
		set data [concat [[irc] ibc-listing [list my-plenaries $channel]]]
		if {$data != ""} {
			foreach plenary $data {
				show-plenary $nick $channel $plenary
			}
		} else {
			[irc] privmsg $nick "No active plenaries"
		}
	}

	proc show-plenary {nick channel plenary} {
		set line "[dict get $plenary plenarychannel]: "
		switch [dict get $plenary type] {
			public {
				append line public
			}
			delegates {
				set delegates [dict get $plenary delegates]
				set plenarychannel [dict get $plenary plenarychannel]
				set delegatessent [[irc] db delegates -from $channel -to $plenarychannel]
				set delegatesleft [- $delegates [llength $delegatessent]]
				append line "$delegates delegates. $delegatesleft left"
			}
		}
		[irc] privmsg $nick $line
	}

	proc show-users {nick channel} {
		set data [[irc] db users $channel]
		[irc] privmsg $nick [join $data]
	}

	proc show-help {nick} {
		[irc] privmsg $nick "http://cntbot.org/docs.html"
	}

	proc show-show-help {nick what} {
		[irc] privmsg $nick [list {!show [channel]} $what [switch $what {
			polls {
				expr {(details <poll-id>)|pending|(list [page])}
			}
			settings {
				expr ""
			}
			plenaries {
				expr ""
			}
			delegates {
				expr {([channel] [nick])}
			}
			users {
				expr ""
			}
			bots {
				expr ""
			}
			bot {
				expr ""
			}
			help {
				expr ""
			}
		}
	}

	# KNOCK

	command knock {
		!knock <channel> (reason|password <password>)
	} {
		variable init
		if {$init} return
		match {{var channel ::hlirc::channel?}
		 		{rest text}} {
			set nick [who]
			if {$channel == [[irc] mainchannel]} {
				# Main channel has extra admission options
				set security [lindex [module config security] 0]
				if {$security == "none"} {
					[irc] invite $channel $nick
					return
				}
			}
			# The channel has admission control or is a plenary
			set user [[irc] nick-user $nick]
			if {$user == ""} {
				[irc] privmsg $nick "Channel $channel has admission control. Please register or log in and try again"
				return
			}
			# Is the user known on this channel
			set user [[irc] nick-user [who]]
			if {[[irc] db user? $channel $user] ||
				[[irc] db delegates -to $channel -user $user] != 0} {
				[irc] invite $channel [who]
				return
			}
			if {$channel == [[irc] mainchannel]} {
				# Main channel has extra admission options
				switch $security {
					password {
						if {$text == ""} {
							[irc] reply "This channel has password-based admission. Please use the form !knock password <password>"
						} elseif {$text == [module config security]} {
							[irc] invite $chan [who]
						} else {
							[irc] privmsg [who] "Wrong password"
						}
					}
					poll {
						if {$text == ""} {
							[irc] reply "This channel requires current members to vote for new members. Please use the form !knock <reason>"
						} else {
							[irc] privmsg $chan "User [who] wants to join this channel. Reason: $text"
							create-voting invite $channel "[who] $text"
						}
					}
				}
			}
		}
	}

	# Module utils

	proc cmd-user-in-channel {who channel} {
		expr {[::hlirc::canonicalize $who] in [[irc] state list users $channel]}
	}

	proc cmd-get-plenary {channel} {
		set myplenary [[irc] db select-plenary $channel]
		if {$myplenary != ""} {
			return $myplenary
		} else {
			set data [[irc] ibc-listing [list plenary $channel]]
			lindex $data 0
		}
	}

	# Get channel users without the bot and ChanServ
	proc channel-users {channel} {
		set users [[irc] state list users $channel]
		struct::set exclude users [[irc] state nick]
		struct::set exclude users chanserv
		return $users
	}


	#### IBC

	proc make-handler {command args body} {
		[irc] ibc-add-handler $command [list apply [list $args $body [namespace current]]]
	}

	make-handler get-data {what args} {
		switch $what {
			plenary {
				lassign $args plenarychannel
				set plenary [[irc] db select-plenary $plenarychannel]
				if {$plenary != ""} {
					dict set plenary bot [[irc] state nick]
					return $plenary
				} else {
					return ""
				}
			}
			my-plenaries {
				lassign $args channel
				set plenaries [[irc] db channel-plenaries $channel]
				struct::list mapfor plenary $plenaries {
					dict merge $plenary [list bot [[irc] state nick]]
				}
			}
		}
	}

	## Remote
	
	proc ibc-show-remote {channel nick what rest} {
		[irc] ibc-send show-remote $channel $nick $what $rest
	}

	make-handler show-remote {channel nick what rest} {
		show-to $nick $what $channel $rest
	}

	## Delegates

	proc ibc-approve-delegate {nick from to} {
		[irc] ibc-send approve-delegate $to $from $nick
	}

	make-handler approve-delegate {to from nick} {
		set plenary [[irc] db select-plenary $to]
		# It's a plenary
		if {$plenary != ""} {
			if {[hlirc::canonicalize $from] ni [dict get $plenary channels]} {
				[irc] privmsg $nick "Your channel isn't allowed for this plenary"
				return
			}
			set current_delegates [llength [[irc] db delegates -from $from -to $to]]
			if {[dict get $plenary type] == "delegates" &&
				$current_delegates >= [dict get $plenary delegates]} {
				[irc] privmsg $nick "Your channel has reached the delegate quota for this plenary"
				return
			}
			do-approve-delegate $nick $from $to
			module after idle ibc-remaining-quota $from $to [expr {[dict get $plenary delegates] - $current_delegates - 1}]
		# It's a protected channel
		} elseif {[[irc] state mode $from i] != ""} {
			[irc] primsg $to "Channel $from sends $nick as delegate"
			create-voting approve-delegate $to "$from $nick"
		# It's a non-protected channel
		} else {
			do-approve-delegate $nick $from $to
		}
	}

	proc ibc-remaining-quota {from to quota} {
		[irc] ibc-send remaining-quota $from $to $quota
	}

	make-handler remaining-quota {from to quota} {
		if {$quota == 0} {
			set quota no
		}
		if {$quota == 1} {
			[irc] privmsg $from "You can send 1 more delegate to plenary $to"
		} else {
			[irc] privmsg $from "You can send $quota more delegates to plenary $to"
		}
	}

	proc ibc-request-delegate {from to} {
		[irc] ibc-send request-delegate $from $to
	}
	
	make-handler request-delegate {from to} {
		[irc] privmsg $from "!!! Channel $to requests a delegate"
	}

	proc ibc-recall-delegate {nick from to reason} {
		[irc] ibc-send recall-delegate $to $nick $from $reason
	}

	make-handler recall-delegate {to nick from reason} {
		set user [[irc] nick-user $nick]
		set delegates [[irc] db delegates -from $from -to $to -user $user]
								
		if {$delegates == ""} {
			module log error "Trying to recall a user who is not a delegate"
			return
		}
		[irc] db remove-delegates -from $from -to $to -user $user
		set timesreceived [llength [[irc] db delegates -to $to -user $user]]
		if {$timesreceived == 0} {
			if {$reason != ""} {
				set msg $reason
			} else {
				set msg "Delegate recalled by $channel"
			}
			remove-delegate $nick $msg
			[irc] kick  $nick $msg
		} else {
			[irc] privmsg $to "Delegate $nick recalled by $channel"
			[irc] privmsg $to "Reason: $reason"
			set channels [[irc] db delegates -user $user -to $to]
			set channels [struct::list mapfor item $channels {[lindex $item 1]}]
			[irc] privmsg $to "$nick is still delegated by the following channels: $channels"
		}
	}

	proc ibc-delegate-approved {from to nick} {
		[irc] ibc-send delegate-approved $from $to $nick
	}

	make-handler delegate-approved {from to nick} {
		set user [[irc] nick-user $nick]
		# Unless we're sending a delegate to our own plenary
		if {[[irc] db select-plenary $to] == ""} {
			[irc] db register-delegate $user $from $to
		}
	}

	proc ibc-delegate-kicked {nick from to reason} {
		[irc] ibc-send delegate-kicked $from $to $nick $reason
	}

	make-handler delegate-kicked {from to nick reason} {
		set delegates [[irc] db delegates -to $to -nick $nick]
		set notifychans [struct:list mapfor delegate $delegates {
			lindex $delegate 2
		}]
		foreach chan $notifychans {
			[irc] privmsg $chan "Delegate $nick was kicked from $to. Reason: $reason"
		}
		[irc] db remove-delegates -nick $nick -to $to
	}

	## Plenaries

	proc ibc-attend-plenary {channel plenarychannel users} {
		[irc] ibc-send attend-plenary $channel $plenarychannel $users
	}

	make-handler attend-plenary {channel plenarychannel users} {
		do-attend-plenary $channel $channel $users
	}

	proc ibc-plenary-dismissed {channel plenarychannel} {
		[irc] ibc-send plenary-dismissed $channel $plenarychannel
	}

	make-handler plenary-dismissed {channel plenarychannel} {
		[irc] db remove-delegates -from $channel -to $plenarychannel
		[irc] ibc-drop-bots-cache
	}

	proc ibc-notify-plenary {hostchannel plenary} {
		foreach channel [dict get $plenary channels] {
			if {$channel == $hostchannel} continue
			[irc] ibc-send notify-plenary $channel $hostchannel \
				[dict get $plenary plenarychannel] \
				[dict get $plenary type] \
				[dict get $plenary delegates] \
				[join [dict get $plenary channels]]
		}
	}

	make-handler notify-plenary {channel hostchannel plenarychannel plenarytype delegates channels} {
		[irc] privmsg $channel "A plenary is assembled with the following channels: $channels"
		if {$plenarytype == "public"} {
			create-voting attend-plenary $channel "$plenarychannel [join $channels]"
		} else {
		}
	}
}
