package provide app-cnt 1.0

namespace eval cnt {

	variable irc
	variable configpath

	proc create-irc {} {
		variable irc
		set irc [::hlirc::new]

		$irc module-eval cnt {}

	}

	proc start-wizard-at-path {configpath} {
		variable irc

		if {$configpath == ""} {
			set configpath ./cnt.conf
		}
		set configpath [file normalize $configpath]

		if {[$irc start-configuration-wizard $configpath]} {
			puts "You can now start the bot as [file dirname $::argv0] $configpath"
			exit
		}
	}

	proc start-wizard {} {
		variable irc
		variable configpath

		create-irc

		print-usage
		puts ""
		puts "If you don't have a configuration file yet it can be created"
		puts "using the first-time setup wizard."
		puts "You can press ^C any time to cancel the wizard."
		puts ""
		puts " Please enter the path to create the configuration file:"
		gets stdin configpath

		start-wizard-at-path $configpath
	}

	proc start-bot {} {
		variable irc
		variable configpath

		create-irc

		if {![file isfile $::argv]} {
			puts "The configuration file you specified doesn't exist."
			puts "Press ^C if it was a mistake or return to start the
first-time setup wizard."
			gets stdin

			start-wizard-at-path $::argv
		}

		signal trap [list sigint sighup] {::cnt::terminate %S}


		set configpath $::argv
		$irc load-config $configpath

		$irc connect
		set ::running 1
		vwait ::running
		$irc log notice "Saving configuration to $configpath"
		$irc save-config $configpath
		$irc destroy
	}

	proc terminate {sig} {
		variable irc
		variable running
		$irc log notice "Received ${sig}. Terminating."
		if {[$irc connected]} {
			$irc quit $sig
		}
		update
		set ::running 0
	}

	proc print-usage {} {
		puts "Usage: [file dirname $::argv0] config-file"
	}

	switch $::argc {
		0 start-wizard
		1 start-bot
		default print-usage
	}
}

