
package require struct::list

# Utils

namespace eval cnt {
	namespace import ::tcl::mathop::*
}

proc ::cnt::lremove {listname item} {
	upvar $listname list
	lsearch -all -inline -not -exact $list $item
}


proc ::cnt::seconds-interval {seconds} {
	if {$seconds % (60 * 60 * 24 * 7) == 0} {
		set seconds [/ $seconds 60 60 24 7]
		set unit week
	} elseif {$seconds % (60 * 60 * 24) == 0} {
		set seconds [/ $seconds 60 60 24]
		set unit day
	} elseif {$seconds % (60 * 60) == 0} {
		set seconds [/ $seconds 60 60]
		set unit hour
	} elseif {$seconds % (60) == 0} {
		set seconds [/ $seconds 60]
		set unit minute
	} else {
		set unit seconds
	}
	set string "$seconds $unit"
	if {$seconds % 10 == 1} {
		return $string
	} else {
		return ${string}s
	}
}

proc ::cnt::named {args defaults} {
	array set defs $defaults
	foreach {key value} $args {
		if {[string index $key 0] != "-"} {
			error "Options start with a dash"
		}
		set key [string range $key 1 end]
		if {![info exists defs($key)]} {
			error "bad option '$key', should be one of: [lsort [array names defs]]"
		}
		set defs($key) $value
	}
	foreach key [array names defs] {
		uplevel [list set $key $defs($key)]
	}
}

proc ::cnt::and {a b} {
	expr {$a && $b}
}

proc ::cnt::nicks-list? {list} {
	struct::list fold [struct::list map $list ::hlirc::nick?] 1 ::cnt::and
}

proc ::cnt::random-string {size {chars "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"}} {
	join [struct::list mapfor _ [struct::list iota $size] {
		set num [expr {int(([string length $chars]) * rand())}]
		string index $chars $num
	}] ""
}
